import requests
from requests import Session
from requests.auth import HTTPBasicAuth
import sys
import xmltodict
import re
import os
import json
import urllib
import AST
import Serviceability
import DIME
from .Logs import Logs
from .CDR import CDR
from . import AXL
from . import Risport
from . import UDS

requests.packages.urllib3.disable_warnings()


class Connect:
    # initialize object set system type and build DB connectors as needed
    def __init__(self, ipaddr=None, username=None, passwd=None, version=None, wsdl=None):
        self.system_type = "cucm"
        self.ipaddr = ipaddr
        self.axl = False
        self.risport = False
        self.auth = HTTPBasicAuth(username, passwd)

        self.uds = UDS.Connect(ipaddr, username, passwd)
        try:
            self.version = self.uds.get("version")['@version']
        except Exception as err:
            print(f"Could not determine CUCM version via UDS: {err}")
            self.version = False

        risport = Risport.Connect(ipaddr, username, passwd)
        self.risport = risport.client

        axl = AXL.Connect(ipaddr, username, passwd, '.'.join(self.version.split('.')[0:2]))
        self.axl = axl.client
        self.add = AXL.Add(self.axl)
        self.do = AXL.Do(self.axl)
        self.get = AXL.Get(self.axl)
        self.list = AXL.List(self.axl)
        self.remove = AXL.Remove(self.axl)
        self.update = AXL.Update(self.axl)

        self.ast = AST.Connect(ipaddr, username, passwd)
        self.cluster = self.ast.cluster
        if self.cluster:
            self.serviceability = Serviceability.Connect(ipaddr, username, passwd, self.cluster['nodes'])
        else:
            self.serviceability = Serviceability.Connect(ipaddr, username, passwd)
        self.dime = DIME.Connect(ipaddr, username, passwd)
        # self.logs = Logs(ipaddr, username, passwd)
        self.cdr = CDR(ipaddr, username, passwd)

        print("Connected.")

    # Function to query SQL via AXL API to CUCM
    def query(self, sql_statement):
        result_dict = []
        row_info = ["", None, None]
        newsql_statement = sql_statement
        while row_info[1] is None or int(row_info[1]) > len(result_dict):
            if re.search("select", sql_statement.lower()):
                execute_sql = "executeSQLQuery"
            else:
                execute_sql = "executeSQLUpdate"

            if row_info[1] is not None and int(row_info[1]) > 0:
                newsql_statement = re.sub("[sS][eE][lL][eE][cC][tT] ",
                                          f"SELECT SKIP {len(result_dict)} LIMIT {row_info[2]} ", sql_statement)

            payload = f"""<?xml version=\"1.0\" encoding=\"UTF-8\"?>
                            <soapenv:Envelope xmlns:soapenv=\"http://schemas.xmlsoap.org/soap/envelope/\" xmlns:ns=\"http://www.cisco.com/AXL/API/10.5\">
                              <soapenv:Header/>
                              <soapenv:Body>
                                <ns:{execute_sql}>
                                  <sql>
                                    {newsql_statement}
                                  </sql>
                                </ns:{execute_sql}>  
                              </soapenv:Body>
                            </soapenv:Envelope>"""
            try:
                # send API request
                response = requests.request("POST", f'https://{self.ipaddr}:8443/axl/',
                                            headers={'Content-Type': 'text/xml'},
                                            auth=self.auth,
                                            data=payload,
                                            verify=False)
            except Exception as err:
                print(f"Error with AXL DB request: {err}", file=sys.stderr)
            else:
                # If request was good then return response values
                if response.status_code == 200:
                    # normalize output to always show array of rows
                    if re.search('ns:executeSQLUpdateResponse', response.text):
                        result_dict.extend([xmltodict.parse(response.text)['soapenv:Envelope']['soapenv:Body'][
                                                'ns:executeSQLUpdateResponse']['return']])
                        break

                    elif xmltodict.parse(response.text)['soapenv:Envelope']['soapenv:Body']['ns:executeSQLQueryResponse'][
                        'return'] is not None:
                        if type(xmltodict.parse(response.text, dict_constructor=dict)['soapenv:Envelope']['soapenv:Body'][
                                    'ns:executeSQLQueryResponse']['return']['row']) is not list:
                            result_dict.extend([xmltodict.parse(response.text, dict_constructor=dict)['soapenv:Envelope'][
                                                    'soapenv:Body']['ns:executeSQLQueryResponse']['return']['row']])
                        else:
                            result_dict.extend(
                                xmltodict.parse(response.text, dict_constructor=dict)['soapenv:Envelope']['soapenv:Body'][
                                    'ns:executeSQLQueryResponse']['return']['row'])

                        if row_info[1] is None or int(row_info[1]) <= len(result_dict):
                            break
                    else:
                        return []

                elif re.search("Query request too large. Total rows matched: ", response.text):
                    row_info = re.sub("[^\d:]", '',
                                      xmltodict.parse(response.text, dict_constructor=dict)['soapenv:Envelope'][
                                          'soapenv:Body']['soapenv:Fault']['faultstring']).split(":")
                    print(f"Multiple queries needed for {row_info[1]} rows.")

                else:
                    print(f'Requests Error: {response}', file=sys.stderr)
                    print(f'\t{sql_statement}', file=sys.stderr)
                    print(f'\t{response.text}', file=sys.stderr)
                    return False

        return result_dict

    # risport query for device registration
    def registration_status(self, device_list=None, return_format="dict"):
        if self.risport is False:
            print("risport client not available.", file=sys.stderr)
            return False
        # If no list given then get all devices to list
        if device_list is None:
            device_list = self.query(f"SELECT name AS device FROM device WHERE tkclass = 1")
        elif type(device_list) == str:
            device_list = [device_list]

        i = 0
        risport_response = {}
        while i < len(device_list):
            # Build and execute the risport request object
            state_info = ''

            # Build Search Criteria
            risport_search_criteria = {
                'MaxReturnedDevices': '1000',
                'DeviceClass': 'Any',
                'Model': '255',
                'Status': 'Any',
                'NodeName': '',
                'SelectBy': 'Name',
                'Protocol': 'Any',
                'DownloadStatus': 'Any',
                'SelectItems': {
                    'item': []
                }
            }
            # One or more specific devices can be retrieved by replacing * with
            # the device name in multiple items
            device_count = 0
            while device_count < 1000 and (device_count + i) < len(device_list):
                if isinstance(device_list[device_count + i], dict):
                    risport_response[device_list[device_count + i]['device']] = {'description': '', 'protocol': '',
                                                                                 'type': '', 'firmware': '', 'Node': '',
                                                                                 'IPAddress': '', 'dirn': '',
                                                                                 'status': 'Never'}
                    risport_search_criteria['SelectItems']['item'].append(
                        {'Item': device_list[device_count + i]['device']})

                elif isinstance(device_list[device_count + i], str):
                    risport_response[device_list[device_count + i]] = {'description': '', 'protocol': '', 'type': '',
                                                                       'firmware': '', 'Node': '', 'IPAddress': '',
                                                                       'dirn': '', 'status': 'Never'}
                    risport_search_criteria['SelectItems']['item'].append({'Item': device_list[device_count + i]})

                device_count = device_count + 1

            i += device_count

            # Get reg statuses
            try:
                resp = self.risport.selectCmDevice(state_info, risport_search_criteria)
            except Fault as err:
                print(f'Zeep error: selectCmDevice: {err}', file=sys.stderr)
                return False

            # format results for easy parsing
            for node in resp['SelectCmDeviceResult']['CmNodes']['item']:
                if node['ReturnCode'] != 'Ok':
                    continue

                for device in node['CmDevices']['item']:
                    risport_response[device['Name']] = {}
                    risport_response[device['Name']]['name'] = device['Name']
                    risport_response[device['Name']]['node'] = node['Name']
                    ipaddresses = device['IPAddress']
                    risport_response[device['Name']]['description'] = device['Description'] if device[
                                                                                                   'Description'] is not None else ''
                    risport_response[device['Name']]['protocol'] = device['Protocol'] if device[
                                                                                             'Protocol'] is not None else ''
                    risport_response[device['Name']]['type'] = device['DeviceClass'] if device[
                                                                                            'DeviceClass'] is not None else ''
                    risport_response[device['Name']]['firmware'] = device['ActiveLoadID'] if device[
                                                                                                 'ActiveLoadID'] is not None else ''
                    risport_response[device['Name']]['IPAddress'] = ipaddresses['item'][0]['IP'] if ipaddresses else ''

                    if device['LinesStatus'] is None:
                        risport_response[device['Name']]['dirn'] = "unknown"
                        risport_response[device['Name']]['status'] = device['Status']

                    else:
                        for x in range(0, len(device['LinesStatus']['item'])):
                            if x == 0:
                                risport_response[device['Name']]['dirn'] = device['LinesStatus']['item'][x][
                                    'DirectoryNumber']
                                risport_response[device['Name']]['status'] = device['LinesStatus']['item'][x]['Status']
                                break

        # if default dict format then return dict
        if return_format == "dict":
            return risport_response

        # else if return format is list then loop through to convert dict based on device name to list of results
        elif return_format == "list":
            risport_list = []
            for item in risport_response:
                if "name" in risport_response[item]:
                    risport_list.append(risport_response[item])

            return risport_list

    def update_pt(self, uuid, partition):
        try:
            dn_return = self.update.Line(uuid=uuid, newRoutePartitionName=partition)

        except Exception as err:
            print(f"Error during DN Update: ", err)
            return False

        else:
            return dn_return



import requests
import json
from .OutputStyle import TextStyle
import time
import urllib

requests.packages.urllib3.disable_warnings()


class Connect:
    access_token = None
    org_id = None
    profile = None
    id = None
    service_url = "https://webexapis.com"
    type = "wbx"

    # store token then fetch users orgId and save
    def __init__(self, access_token=None, type="wbx"):
        if not access_token:
            TextStyle.error("Must pass API token for Admin API")
        else:
            self.access_token = access_token

            try:
                self.profile = self.get("/v1/people/me?callingData=true")
                self.od = self.profile['id']
                self.org = self.profile['orgId']

            except Exception as err:
                raise Exception(f"Error with Token or getting Org")

    # parses out paging from response headers
    @staticmethod
    def find_next_page(call_response):
        if 'Link' in call_response.headers:
            return call_response.headers['Link'].split(";")[0].strip("<>")
        else:
            return False

    # basic REST GET
    def get(self, uri, data=None, limit=50000):
        # URL encode parameters for appending to URI if provided
        if data:
            data = f"?{urllib.urlencode(data)}"
        else:
            data = ""

        headers = {"Accept": "application/json",
                   "Authorization": f"Bearer {self.access_token}"}  # set auth header

        uri = f"{self.service_url}/{uri.strip('/')}{data}"
        return_values = {'items': []}
        # Loop while true to handled paged results and throttling
        while True:
            try:
                # send request to Webex
                response = requests.get(uri, headers=headers, verify=False)

            except Exception as err:
                raise Exception(f"Error sending Webex REST GET - {uri} {err}")

            else:
                if 200 <= response.status_code <= 300 and (response.headers['Content-Type'].find("json") > -1):
                    # create or add to return dict
                    response_data = response.json()
                    if 'items' in response_data:
                        return_values['items'].extend(response_data['items'])
                    else:
                        return response_data

                    # check for additional and loop until max (limit minimum is 50)
                    if len(return_values['items']) >= limit or self.find_next_page(response) is False:
                        return return_values
                    else:
                        uri = self.find_next_page(response)
                elif 200 <= response.status_code < 300:
                    return True
                elif response.status_code == 429:  # handle throttling
                    TextStyle.warning(f"Request to {uri} throttled.  Doing it again in a bit.")
                    time.sleep(int(response.headers['Retry-After']) + 1)

                elif response.status_code == 401:  # handle auth issues
                    TextStyle.error("Webex Admin token invalid.  Generate new token")
                    return False
                else:  # catch generic failures
                    TextStyle.error(f"Error sending GET to Webex - {uri}{data} {response}{response.text}")
                    return False

    # basic REST DELETE
    def delete(self, uri):
        headers = {"Accept": "application/json",
                   "Authorization": f"Bearer {self.access_token}"}

        while True:  # loop for throttling
            try:
                response = requests.delete(f"{self.service_url}/{uri.strip('/')}", headers=headers, verify=False)

            except Exception as err:
                raise Exception(f"Error sending Webex REST DELETE - {uri} {err}")

            else:
                if response.status_code == 426:  # handle throttling
                    time.sleep(int(response.headers['Retry-After']) + 1)
                elif response.status_code == 200 and (response.headers['Content-Type'].find("json") > -1):
                    return json.loads(response.text)
                else:
                    TextStyle.error(f"Error sending DELETE to Webex - {uri} {response}{response.text}")
                    return False

    # basic REST POST
    def post(self, uri, data):
        headers = {"Accept": "application/json",
                   "Content-Type": "application/json",
                   "Authorization": f"Bearer {self.access_token}"}

        while True:  # loop for throttling
            try:
                response = requests.post(f"{self.service_url}/{uri.strip('/')}",
                                         headers=headers,
                                         data=json.dumps(data),
                                         verify=False)

            except Exception as err:
                raise Exception(f"Error sending Webex REST POST - {uri} {err}")

            else:
                if response.status_code == 426:  # handle throttling
                    time.sleep(int(response.headers['Retry-After']) + 1)
                elif response.status_code == 200 and (response.headers['Content-Type'].find("json") > -1):
                    return json.loads(response.text)
                else:
                    TextStyle.error(f"Error sending POST to Webex - {uri} {response}{response.text}")
                    return False

    # basic REST PUT
    def put(self, uri, data):
        headers = {"Accept": "application/json",
                   "Content-Type": "application/json",
                   "Authorization": f"Bearer {self.access_token}"}

        while True:  # loop for throttling\
            try:
                response = requests.put(f"{self.service_url}/{uri.strip('/')}",
                                        headers=headers,
                                        data=json.dumps(data),
                                        verify=False)

            except Exception as err:
                raise Exception(f"Error sending Webex REST PUT - {uri} {err}")

            else:
                if response.status_code == 426:  # handle throttling
                    time.sleep(int(response.headers['Retry-After']) + 1)
                elif response.status_code == 200 and (response.headers['Content-Type'].find("json") > -1):
                    return json.loads(response.text)
                else:
                    TextStyle.error(f"Error sending PUT to Webex - {uri} {response}{response.text}")
                    return False

    # basic REST patch
    def patch(self, uri, data):
        headers = {"Accept": "application/json",
                   "Content-Type": "application/json",
                   "Authorization": f"Bearer {self.access_token}"}

        while True:  # loop for throttling
            try:
                response = requests.patch(f"{self.service_url}/{uri.strip('/')}",
                                          headers=headers,
                                          data=json.dumps(data),
                                          verify=False)

            except Exception as err:
                raise Exception(f"Error sending Webex REST PATCH - {uri} {err}")

            else:
                if response.status_code == 426:  # handle throttling
                    time.sleep(int(response.headers['Retry-After']) + 1)
                elif response.status_code == 200 and (response.headers['Content-Type'].find("json") > -1):
                    return json.loads(response.text)
                else:
                    TextStyle.error(f"Error sending PATCH to Webex - {uri} {response}{response.text}")
                    return False



import sys
import requests
from zeep.transports import Transport
from zeep import Client, Settings
from zeep.helpers import serialize_object
import re

requests.packages.urllib3.disable_warnings()


class Connect:
    app_creds = None
    version = None
    nodes = {}
    service_list = []

    def __init__(self, ipaddr=None, username=None, passwd=None, nodes=None):
        print("Connecting to CM Servicability...")
        if ipaddr is None:
            print("Must declare connection info: ", file=sys.stderr)
            print("Collab.CMServiceManager(ipaddr='x.x.x.x', app_creds=('appAdmin', 'SECRET')",
                  file=sys.stderr)
        else:
            self.ipaddr = ipaddr
            if username is None or passwd is None:
                print("Must declare connection info: ", file=sys.stderr)
                print("Collab.CMServiceManager(ipaddr='x.x.x.x', app_creds=('appAdmin', 'SECRET')",
                      file=sys.stderr)
            else:
                self.app_creds = requests.auth.HTTPBasicAuth(username, passwd)
                print("Getting node capabilities...")
                if nodes is None:
                    nodes = [ipaddr]
                for node in nodes:
                    self.discover_node(node)
        print("... Done")

    # discover platform components for nodes and available services and store
    def discover_node(self, ipaddr):
        # set WSDL file from URL
        controlcenter_wsdl = f'https://{ipaddr}:8443/controlcenterservice2/services/ControlCenterServices?wsdl'

        # create a SOAP client session
        session = requests.Session()

        # avoid certificate verification by default and setup session
        session.verify = False
        session.auth = self.app_creds
        transport = Transport(session=session, timeout=10)
        settings = Settings(strict=False, xml_huge_tree=True)

        # Create client and soap service
        client_controlcenter = Client(controlcenter_wsdl, settings=settings,
                                      transport=transport)  # , plugins = plugin )
        try:
            soap_controlcenter = client_controlcenter.create_service(
                '{http://schemas.cisco.com/ast/soap}ControlCenterServicesBinding',
                controlcenter_wsdl)
        except Exception as err:
            print(f"Error creating SOAP Connector: {err}", file=sys.stderr)

        else:
            try:
                server_list = soap_controlcenter.getProductInformationList(ServiceInfo="")

            except Exception as err:
                print(f"Error getting server list and info: {err}", file=sys.stderr)
                return False

            else:
                self.version = server_list['ActiveServerVersion']
                if ipaddr not in list(self.nodes.keys()):
                    self.nodes[ipaddr] = { 'name': server_list['PrimaryNode'], 'soap': soap_controlcenter }

                if self.service_list == []:
                    [self.service_list.append(i['ServiceName']) for i in server_list['Services']['item']]

                return True

    # convert list return into a dict based on the service name
    @staticmethod
    def list_to_dict(return_list):
        results = {}
        for service in return_list['ServiceInfoList']['item']:
            try:
                results[service['ServiceName']] = serialize_object(service, dict)
            except:
                results[service['ServiceName']] = service

        return results

    # check requested service against actual services and try to resolve
    def verify_service_name(self, service):
        if service in self.service_list:
            return service

        else:
            for i in self.service_list:
                if re.sub("[^a-z]","",service.lower().replace("cisco","")) == re.sub("[^a-z]","",i.lower().replace("cisco","")):
                    return i

        print('Cannot find service in list', file=sys.stderr)
        return False

    # return service status of all services
    def status(self, service_list=None, nodes=None):
        # if nodes not specified then assume all nodes in cluster
        if nodes is None:
            nodes = self.nodes

        elif isinstance(nodes, str):
            nodes = [nodes]

        if service_list is not None:
            if isinstance(service_list, list):
                # loop through the requested service list and try and correct naming
                x = 0
                while x < len(service_list):
                    service = self.verify_service_name(service_list[x])
                    if service is not False:
                        service_list[x] = self.verify_service_name(service_list[x])
                    else:
                        print(f"Unknown Service {service_list[x]}", file=sys.stderr)
                        service_list.pop(x)
                    x += 1
            else:
                service_list = [self.verify_service_name(service_list)]

        else:
            service_list = ""

        results = {}
        for node in nodes:
            try:
                service_status = self.nodes[node]['soap'].soapGetServiceStatus(ServiceStatus=service_list)

            except Exception as err:
                print(f"Error getting service status: {err}", file=sys.stderr)

            else:
                results[node] = self.list_to_dict(service_status)

        return results

    # actions based on service list n node list
    def service(self, action, service_list, nodes=None):
        activate = False
        if isinstance(service_list, str):
            service_list = [service_list]

        # if none list is blank then assume all nodes
        if nodes is None:
            nodes = self.nodes
        elif isinstance(nodes, str):
            nodes = [nodes]

        if action.lower() == "status":
            return self.get_service_status(service_list, nodes)
        elif action.lower() == "start":
            action = "Start"
        elif action.lower() == "restart":
            action = "Restart"
        elif action.lower() == "stop":
            action = "Stop"
        elif action.lower() == "deactivate":
            activate = True
            action = "UnDeploy"
        elif action.lower() == "activate":
            activate = True
            action = "Deploy"
        else:
            print(f"\tUnknown action command - {action}", file=sys.stderr)
            return False

        # loop through service list and try and confirm/correct naming
        x = 0
        while x < len(service_list):
            service = self.verify_service_name(service_list[x])
            if service is not False:
                service_list[x] = self.verify_service_name(service_list[x])
            else:
                print(f"Unknown Service: {service_list[x]}", file=sys.stderr)
                service_list.pop(x)
            x += 1

        # if no valid services then skip fetching
        if len(service_list) == 0:
            print(f"No valid services to check.", file=sys.stderr)
            return False
        else:
            results = {}
            # loop services on nodes
            for node in nodes:
                print(f"\t{action}ing {service_list} on {node}...")
                try:
                    if activate:
                        result = self.nodes[node]['soap'].soapDoServiceDeployment(
                            DeploymentServiceRequest={'NodeName': node, 'DeployType': action,
                                                      'ServiceList': service_list})
                    else:
                        result = self.nodes[node]['soap'].soapDoControlServices(
                            ControlServiceRequest={'NodeName': node, 'ControlType': action, 'ServiceList': service_list})

                except Exception as err:
                    print(f"\tError {action}ing {service_list}: {err}", file=sys.stderr)

                else:
                    # reformat and append results
                    results[node] = self.list_to_dict(result)

            return results

    def start(self, service_list, nodes=None):
        return self.service("start", service_list, nodes)

    def stop(self, service_list, nodes=None):
        return self.service("stop", service_list, nodes)

    def restart(self, service_list, nodes=None):
        return self.service("restart", service_list, nodes)

    def activate(self, service_list, nodes=None):
        return self.service("activate", service_list, nodes)

    def deactivate(self, service_list, nodes=None):
        return self.service("deactivate", service_list, nodes)

    def find_tftp(self):
        print("Finding TFTP servers...")
        tftp_servers = self.status("Cisco Tftp")

        server_list = []
        for server in tftp_servers:
            if tftp_servers[server]['Cisco Tftp']['ReasonCodeString'].find("Service Not Activated") < 0:
                server_list.append(server)

        return server_list

    def find_moh(self):
        print("Finding Media Streaming servers...")
        moh_servers = self.status("Cisco IP Voice Media Streaming App")

        server_list = []
        for server in moh_servers:
            if moh_servers[server]['Cisco IP Voice Media Streaming App']['ReasonCodeString'].find("Service Not Activated") < 0:
                server_list.append(server)

        return server_list



import requests
from requests import Session
from requests.auth import HTTPBasicAuth
import sys
import os
from zeep.transports import Transport
from zeep import Client, Settings

requests.packages.urllib3.disable_warnings()


class Connect:
    def __init__(self, ipaddr, username, passwd, version="14.0", wsdl=None):
        # if type= cucm then set username/password for AXL connection
        if ipaddr is None or passwd is None or username is None:
            print(f'Usage: CollabConnector.AXL("ipaddr", "admin", "password", version="12.0", wsdl="./AXL/AXLAPI.wsdl")',
                  file=sys.stderr)
        else:
            self.username = username

            if wsdl:
                wsdl = wsdl
            elif version and int(version.split(".")[0]) < 10:
                wsdl = os.path.join(os.path.dirname(__file__), 'schema', '10.0', 'AXLAPI.wsdl')
            elif version:
                wsdl = os.path.join(os.path.dirname(__file__), 'schema', version, 'AXLAPI.wsdl')

            # create a SOAP client session
            session = Session()

            # avoid certificate verification by default and setup session
            session.verify = False
            session.auth = HTTPBasicAuth(username, passwd)
            transport = Transport(session=session, timeout=10)
            settings = Settings(strict=False, xml_huge_tree=True)

            # If WSDL file specified then create AXL SOAP connector
            if wsdl is not None:
                # Create the Zeep client with the specified settings
                client_axl = Client(wsdl, settings=settings, transport=transport)  # ,plugins = plugin )
                # Create the Zeep service binding to AXL at the specified CUCM
                try:
                    self.client = client_axl.create_service(
                        '{http://www.cisco.com/AXLAPIService/}AXLAPIBinding',
                        f'https://{ipaddr}:8443/axl/')

                except Exception as err:
                    print(f"SOAP/AXL Error could not create service: {err}", file=sys.stderr)
                    self.client = False